const { ActivityFlags, Util, Constants, Base, Emoji } = require('discord.js');

class Presence extends Base {
  constructor(client, data = {}) {
    super(client);
    this.userId = data.user.id;
    this.guild = data.guild || null;
    this._patch(data);
  }

  get user() {
    return this.client.users.resolve(this.userId);
  }

  get member() {
    return this.guild.members.resolve(this.userId);
  }

  _patch(data) {
    this.status = data.status || this.status || 'offline';
    this.activities = data.activities.map(activity => new Activity(this, activity)) || [];
    this.clientStatus = data.client_status || null;
    return this;
  }

  _clone() {
    const clone = Object.assign(Object.create(this), this);
    clone.activities = this.activities.map(activity => activity._clone());
    return clone;
  }

  equals(presence) {
    return (
      this === presence ||
      (presence &&
        this.status === presence.status &&
        this.activities.length === presence.activities.length &&
        this.activities.every((activity, index) => activity.equals(presence.activities[index])) &&
        this.clientStatus.web === presence.clientStatus.web &&
        this.clientStatus.mobile === presence.clientStatus.mobile &&
        this.clientStatus.desktop === presence.clientStatus.desktop)
    );
  }

  toJSON() {
    return Util.flatten(this);
  }
}

class Activity {
  constructor(presence, data) {
    Object.defineProperty(this, 'presence', { value: presence });
    this.id = data.id;
    this.name = data.name;
    this.type = typeof data.type === 'number' ? ActivityTypes[data.type] : data.type;
    this.url = data.url || null;
    this.details = data.details || null;
    this.state = data.state || null;
    this.applicationId = data.application_id || null;
    this.timestamps = data.timestamps
      ? {
          start: data.timestamps.start ? new Date(Number(data.timestamps.start)) : null,
          end: data.timestamps.end ? new Date(Number(data.timestamps.end)) : null,
        }
      : null;
    this.syncId = data.sync_id || null;
    this.platform = data.platform || null;
    this.party = data.party || null;
    this.assets = data.assets ? new RichPresenceAssets(this, data.assets) : null;
    this.flags = new ActivityFlags(data.flags).freeze();
    this.emoji = data.emoji ? new Emoji(presence.client, data.emoji) : null;
    this.sessionId = data.session_id || null;
    this.buttons = data.buttons || [];
    this.createdTimestamp = new Date(data.created_at).getTime();
  }

  equals(activity) {
    return (
      this === activity ||
      (activity &&
        this.name === activity.name &&
        this.type === activity.type &&
        this.url === activity.url &&
        this.state === activity.state &&
        this.details === activity.details)
    );
  }

  get createdAt() {
    return new Date(this.createdTimestamp);
  }

  toString() {
    return this.name;
  }

  _clone() {
    return Object.assign(Object.create(this), this);
  }
}

class RichPresenceAssets {
  constructor(activity, assets) {
    Object.defineProperty(this, 'activity', { value: activity });
    this.largeText = assets.large_text || null;
    this.smallText = assets.small_text || null;
    this.largeImage = assets.large_image || null;
    this.smallImage = assets.small_image || null;
  }

  smallImageURL({ format, size } = {}) {
    return (
      this.smallImage &&
      this.activity.presence.client.rest.cdn.AppAsset(this.activity.applicationId, this.smallImage, {
        format,
        size,
      })
    );
  }

  largeImageURL({ format, size } = {}) {
    if (!this.largeImage) return null;
    if (/^spotify:/.test(this.largeImage)) {
      return `https://i.scdn.co/image/${this.largeImage.slice(8)}`;
    } else if (/^twitch:/.test(this.largeImage)) {
      return `https://static-cdn.jtvnw.net/previews-ttv/live_user_${this.largeImage.slice(7)}.png`;
    }
    return this.activity.presence.client.rest.cdn.AppAsset(this.activity.applicationId, this.largeImage, {
      format,
      size,
    });
  }
}

exports.Presence = Presence;
exports.Activity = Activity;
exports.RichPresenceAssets = RichPresenceAssets;
